/* manex1.c */

/* Example in man page for pthread_cond, slightly adjusted  *
 *                                                          */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <limits.h>
#include <pthread.h>

#define MAXCONS 8     // no more than will fit on a terminal line

char state[MAXCONS+1];

int x,y;
pthread_mutex_t mut = PTHREAD_MUTEX_INITIALIZER;
pthread_cond_t cond = PTHREAD_COND_INITIALIZER;

void thrdPrint(int thrdindx, char * msg)
{
    int i;

    for (i = 0; i < thrdindx; i++)
	    printf("%c    ", state[i]);
    printf(msg, thrdindx);
	for (i = thrdindx+1; i <= MAXCONS; i++)
	    printf("%c    ", state[i]);
	printf("\n");
}

void * producer(void * arg)
{
	state[0] = '|';

    for (;;) {
        pthread_mutex_lock(&mut); 
        thrdPrint(0, "<    "); 
        x = rand(); y = rand(); 

        if (x > y) {
	        thrdPrint(0, "xGTy ");
            pthread_cond_broadcast(&cond);
	    }
        thrdPrint(0, ">    ");
        pthread_mutex_unlock(&mut);

	    sleep (1 + rand()%(MAXCONS+2));
	}
}
////

void * consumer(void * arg)
{
	int nc;    // this is the consumer/thread number - a local!

	nc = *((int * ) arg);
	state[nc] = '|';

    for (;;) {
        pthread_mutex_lock(&mut);
        thrdPrint(nc, "<    ");
        while (x <= y) {
		    thrdPrint(nc, ">w   ");
		    state[nc] = '-';
            pthread_cond_wait(&cond, &mut);
		    state[nc] = '|';
            thrdPrint(nc, "w<   ");		   
		}
        thrdPrint(nc, "xGTy ");
        thrdPrint(nc, ">    ");
        pthread_mutex_unlock(&mut);

	    sleep (nc + rand()%nc);
	}
}

int main(int argc, char *argv[])
{
	pthread_t thrd[1+MAXCONS];
	int i;
	int indx[1+MAXCONS];

    srand((unsigned int) time(NULL));

    for (i = 0; i <= MAXCONS; i++)  // indx[i] is used to pass
	    indx[i] = i;                // to each thread its index
                                    // 0=prod, 1...MAXCONS = cons
   	pthread_create(&thrd[0],  NULL, producer, 
                   (void *) &indx[0]);  
	for (i = 1; i <= MAXCONS; i++)
		pthread_create(&thrd[i], NULL, consumer, 
		               (void *) &indx[i]);

    pthread_exit(NULL);
	exit(0);
}
