/* manex2.c */

/* Example in man page for pthread_cond, slightly adjusted *
 *                                                         */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <limits.h>
#include <pthread.h>
#include <sys/time.h>
#include <errno.h>     // for ETIMEDOUT

#define MAXCONS 4     // no more than will fit on a terminal line

int x,y;
pthread_mutex_t mut = PTHREAD_MUTEX_INITIALIZER;
pthread_cond_t cond = PTHREAD_COND_INITIALIZER;

void thrdPrint(int thrdindx, char * msg)
{
    int i;

    for (i = 0; i < thrdindx; i++)
	    printf(".    ");
    printf(msg, thrdindx);
}

void * producer(void * arg)
{
    for (;;) {
        pthread_mutex_lock(&mut); 
        thrdPrint(0, "<\n"); 
        x = rand(); y = rand(); 

        if (x > y) {
            pthread_cond_broadcast(&cond);
	        thrdPrint(0, "xGTy\n");
	    }
        thrdPrint(0, ">\n");
        pthread_mutex_unlock(&mut);

	    sleep (1 + rand()%6);
	}
}
////

void * consumer(void * arg)
{
	int nc;    // this is the consumer/thread number

    struct timeval now;
    struct timespec timeout;
    int retcode;

	nc = *((int * ) arg);

    for (;;) {
        pthread_mutex_lock(&mut);
        thrdPrint(nc, "<\n");

        gettimeofday(&now, NULL);
        timeout.tv_sec = now.tv_sec + rand()%6;
        timeout.tv_nsec = now.tv_usec * 1000;

	    retcode = 0;
	    while ( x <= y && retcode != ETIMEDOUT ) {
		    thrdPrint(nc, ">\n");		   
            retcode = 
			   pthread_cond_timedwait(&cond, &mut, &timeout);
            thrdPrint(nc, "<\n");		   
		}
	    if (retcode == ETIMEDOUT)
		    thrdPrint(nc, "tout\n");
        else 
		    thrdPrint(nc, "xGTy\n");
        thrdPrint(nc, ">\n");
        pthread_mutex_unlock(&mut);

	    sleep (nc + rand()%nc);
	}
}
////

int main(int argc, char *argv[])
{
	pthread_t thrd[1+MAXCONS];
	int i;
	int indx[1+MAXCONS];

    srand((unsigned int) time(NULL));

    for (i = 0; i <= MAXCONS; i++)  // indx[i] is used to pass
	    indx[i] = i;                // to each thread its index
                                    // 0=prod, 1...MAXCONS = cons
   	pthread_create(&thrd[0],  NULL, producer, &indx[0]);  
	for (i = 1; i <= MAXCONS; i++)
		pthread_create(&thrd[i], NULL, consumer, 
		               (void *) &indx[i]);

    pthread_exit(NULL);
	exit(0);
}