/* philosophers.c */

/* Dining Philosophers */

#include <stdio.h>
#include <pthread.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>

#define N   5
#define LEFT_n	n
#define RIGHT_n	(n + 1) % N

#define THINK   0
#define HUNGRY  1
#define EAT     2
#define PRINT_ST(state)	\
	( (state == THINK)  ? "Think" : \
	  (state == EAT)    ? "Eat" : \
	  (state == HUNGRY) ? "Hungry" : \
	                      "no such state")


pthread_mutex_t theFork[N], mutex;
pthread_cond_t state_changed;
int state[N];

void setPhState(int phil, int st)
{
	pthread_mutex_lock(&mutex);
	state[phil] = st;
	pthread_mutex_unlock(&mutex);
	// signal show_state thread
	pthread_cond_broadcast(&state_changed);	
//	pthread_cond_signal(&state_changed);		// works too
}

void think(int n)
{
	setPhState(n,THINK);
#ifndef debug
	sleep(1+(int)(3.0*rand() / (RAND_MAX+1.0)));
#endif
}
////

void eat(int n)
{
	setPhState(n,EAT);
#ifndef debug	
	sleep(1+(int)(3.0*rand() / (RAND_MAX+1.0)));
#endif
}

void take_forks(int n)
// Get both forks for philosopher n (block until both available)
{
	setPhState(n,HUNGRY);
	for (;;) {
		pthread_mutex_lock(&theFork[LEFT_n]);
		if (pthread_mutex_trylock(&theFork[RIGHT_n]) != EBUSY)
			break;
		pthread_mutex_unlock(&theFork[LEFT_n]);

		pthread_mutex_lock(&theFork[RIGHT_n]);
		if (pthread_mutex_trylock(&theFork[LEFT_n]) != EBUSY)
			break;
		pthread_mutex_unlock(&theFork[RIGHT_n]);
	}
}

void put_forks(int n)
{
	setPhState(n, THINK);
//	previous line is necessary, even though this will be done also 
//	when main loop restarts; otherwise, the unlocks below could set a
//	neighbour to state EAT, and show_state() would display this phil's 
//	state as EAT
	pthread_mutex_unlock(&theFork[RIGHT_n]);
	pthread_mutex_unlock(&theFork[LEFT_n]);
}
////

void * philosopher(void * x)
{
	int n = *(int *) x;
	for (;;) {
		think(n);
		take_forks(n);
		eat(n);
		put_forks(n);
	}
}

void * show_state(void * x)
{
	int n;
#ifdef debug
	int times = 0;
#endif

	for (n = 0; n < N; n++) {
		printf("%2d        ", n);
	}
	printf("\n");
	
	for (;;) {
		pthread_mutex_lock(&mutex);
		pthread_cond_wait(&state_changed, &mutex);
		for (n = 0; n < N; n++) {
#ifndef debug
			printf("%-10s", PRINT_ST(state[n]));
		}
		printf("\n");
#else
			if (state[n] == EAT && state[RIGHT_n] == EAT)
					printf("Error at iter. %d\n", times);
		}
		++times;
		if (times % 2000 == 0)
			printf("Iter %d\n", times);
#endif
		pthread_mutex_unlock(&mutex);
	}
}
////

int main(int argc,char *argv[])
{
	pthread_t tid[N+1];
	int targ[N+1];
	int i;

	pthread_mutex_init(&mutex, NULL);
	pthread_cond_init(&state_changed, NULL);
	for (i = 0; i < N; i++)
		pthread_mutex_init(&theFork[i], NULL);

    srand((unsigned int) time(NULL));
   
	for (i = 0; i < N; i++) {
		targ[i] = i;
		pthread_create(&tid[i], NULL, philosopher, (void *) &targ[i]);
	}
	pthread_create(&tid[N], NULL, show_state, NULL);
	pthread_join(tid[N], NULL);
	exit(0);
}
