/* synqueue.c */

/* ADT queue, synchronized nel senso del Java, cioe`        *
 * con operazioni atomiche. Si tratta quindi di un monitor. *
 * 															*
 * Funzionamento:											*
 * 1. chi non trova dati, fa wait su for_data				*
 * 2. chi produce un dato, sveglia (broadcast) chi e` in wait*
 * 3. quindi chi si sveglia da wait potrebbe non trovare	*
 *    dati, se un altro svegliato li prende prima, e deve	*
 *    percio` riprovare (1) (con un while)					*/

#include <stdlib.h>
#include <pthread.h>
#include "synqueue.h"

int q_init(syn_queue_t * q, int size)
{
	q->num = q->head = q->tail = 0;
	q->size = size;
	pthread_mutex_init(&q->mutex, NULL);
	pthread_cond_init(&q->for_space, NULL);
	pthread_cond_init(&q->for_data, NULL);
	q->data = (long *) malloc(sizeof(long) * size);
	return (q->data == NULL) ? 1 : 0;
}
////

void q_free(syn_queue_t * q)
{
  if (q->data != NULL)
    free(q->data);
}

int q_get(syn_queue_t * q, long * val, int which_thrd)
{
	pthread_mutex_lock(&q->mutex);
	while (q->num == 0)   // try to change while/if -> error!
		pthread_cond_wait(&q->for_data, &q->mutex);
	*val = q->data[q->tail];
	show_val(which_thrd, "get", *val);
	q->tail = (q->tail + 1) % q->size;
	q->num--;
	pthread_cond_broadcast(&q->for_space);
	pthread_mutex_unlock(&q->mutex);	
	return 0;
}

int q_put(syn_queue_t * q, long val)
{
	pthread_mutex_lock(&q->mutex);
	while (q->num == q->size)
		pthread_cond_wait(&q->for_space, &q->mutex);
	show_val(0, "put", val);
	q->data[q->head] = val;
	q->head = (q->head + 1) % q->size;
	q->num++;
	pthread_cond_broadcast(&q->for_data);
	pthread_mutex_unlock(&q->mutex);
	return 0;
}
