/* synqueue1.c */

/* ADT queue, synchronized nel senso del Java, cioe`        *
 * con operazioni atomiche. Si tratta quindi di un monitor. *
 */

#include <stdlib.h>
#include <pthread.h>

#include "synqueue.h"

int q_init(syn_queue_t * q, int size)
{
	q->num = q->head = q->tail = 0;
	q->size = size;
	q->data = (long *) malloc(sizeof(long) * size);
	pthread_mutex_init(&q->mutex, NULL);
	pthread_cond_init(&q->for_space, NULL);
	pthread_cond_init(&q->for_data, NULL);
	q->n_wait_dt = q->n_wait_sp = 0;
	return (q->data == NULL) ? 1 : 0;
}

void q_free(syn_queue_t * q)
{
  if (q->data != NULL)
    free(q->data);
}
////

int q_get(syn_queue_t * q, long * val, int which_thrd)
{
	pthread_mutex_lock(&q->mutex);
	if (q->num == 0) {
		q->n_wait_dt++;
		pthread_cond_wait(&q->for_data, &q->mutex);
	}
	*val = q->data[q->tail];
	show_val(which_thrd, "get", *val);
	q->tail = (q->tail + 1) % q->size;
	if ((q->num-- == q->size) &&  // not useless, although we count threads waiting for space
	    (q->n_wait_sp > 0) ) {
		q->n_wait_sp--;
		pthread_cond_signal(&q->for_space);
	}								  // after signal also need unlock()
	pthread_mutex_unlock(&q->mutex);  //   or threads blocked on wait
	return 0;						  //   will not proceed
}                                       

int q_put(syn_queue_t * q, long val)
{
	pthread_mutex_lock(&q->mutex);
	if (q->num == q->size) {
		q->n_wait_sp++;
		pthread_cond_wait(&q->for_space, &q->mutex);
	}
	q->data[q->head] = val;
	q->head = (q->head + 1) % q->size;
	show_val(0, "put", val);
	if ((q->num++ == 0) &&		// not useless, although we count threads waiting for space
		(q->n_wait_dt > 0) ) {
		q->n_wait_dt--;
		pthread_cond_signal(&q->for_data);
	}								  // after signal also need unlock() 
	pthread_mutex_unlock(&q->mutex);  //   or threads blocked on wait
	return 0;						  //   will not proceed
}									  
